from django.conf import settings

defaults = {
    'DEBUG': False,
    'O365_MAIL_CLIENT_ID': None,
    'O365_MAIL_CLIENT_SECRET': None,
    'O365_MAIL_TENANT_ID': None,
    'O365_MAIL_MAILBOX_KWARGS': {},
    'O365_MAIL_REPLACE_LINE_ENDINGS': True,
    'O365_MAIL_SAVE_TO_SENT': False,
    'O365_ACTUALLY_SEND_IN_DEBUG': False,
}

for key, value in list(defaults.items()):
    if not hasattr(settings, key):
        globals()[key] = value
    else:
        globals()[key] = getattr(settings, key)



from .file_base import BaseConverter
import io


class TupleToFileObject(BaseConverter):
    def __init__(self, obj):
        assert isinstance(obj, tuple) == True, "Object must be an instance of tuple!"
        self.obj = obj

    def get_file(self):
        """Returns a file-like object, in this case a BytesIO object."""

        # A tuple instance has three parts: filename, byte-encoded contents, mimetype
        self.filename = self.obj[0]
        self.content = self.obj[1].encode() if isinstance(self.obj[1], str) else self.obj[1]
        self.mimetype = self.obj[2]

        return io.BytesIO(self.content)
    
    def get_filename(self):
        return self.filename

    def is_inline(self):
        return False

    def get_content_id(self):
        return None



import re
from django.core.mail.message import EmailMultiAlternatives
from . import settings

from email.mime.base import MIMEBase
from .file_mimebase import MIMEObjectToFileObject
from .file_tuple import TupleToFileObject

def string_contains_html(string):
    return (re.search(r'<.+?>', string) is not None)

def get_html_message(message):
    """
    Returns None if the email is plain text only, otherwise returns the HTML message
    """
    if type(message) == EmailMultiAlternatives:
        if message.alternatives:
            for alt in message.alternatives:
                if alt[1] == 'text/html':
                    return alt[0]
    return None

def get_message_body(message):
    """
    Function to get the body for the email message, depending on settings
    """
    html_body = get_html_message(message)
    if not html_body:
        if settings.O365_MAIL_REPLACE_LINE_ENDINGS and not string_contains_html(message.body):
            html_body = message.body.replace("\n", "<br />\n")
        else:
            html_body = message.body
    return html_body

def get_name_and_email(address):
    """
    Function to get name and email from addresses like
    Company <contact@company.com>, returned as tuple (name, email)
    """
    custom_sender_name = re.search(r'^([^<>]+)\s<([^<>]+)>$', address)
    if custom_sender_name:
        return custom_sender_name.group(1), custom_sender_name.group(2)
    else:
        return "", address

def get_converter(attachment):
    converter = None

    # Determine converter
    if isinstance(attachment, tuple):
        converter = TupleToFileObject
    elif isinstance(attachment, MIMEBase):
        converter = MIMEObjectToFileObject
    else:
        raise Exception("Invalid attachment type!")

    return converter


from django.core.mail.backends.base import BaseEmailBackend
import threading
import O365

from . import settings
from . import util

import logging
from .o365_logger import SimpleErrorHandler # Handles auth exceptions!


"""
A wrapper that manages the O365 API for sending emails.
Uses an identity (auth_flow_type == 'credentials').
See https://docs.microsoft.com/en-us/graph/auth-v2-service?context=graph%2Fapi%2F1.0&view=graph-rest-1.0 for more details.
"""


class O365EmailBackend(BaseEmailBackend):
    def __init__(self, client_id=None, client_secret=None, tenant_id=None,
                 fail_silently=False, **kwargs):
        super().__init__(fail_silently=fail_silently)
        self.client_id = client_id or settings.O365_MAIL_CLIENT_ID
        self.client_secret = client_secret or settings.O365_MAIL_CLIENT_SECRET
        self.tenant_id = tenant_id or settings.O365_MAIL_TENANT_ID

        self.mailbox = None
        
        # Handle exceptions that come from authentication (Only errors)
        # This is needed because O365 does not raise Exceptions, it only logs them.
        self.log_handler = SimpleErrorHandler()
        log = logging.getLogger('O365')
        log.addHandler(self.log_handler)

        self._lock = threading.RLock()

    def open(self):
        """
        Ensure an open connection to the email server. Return whether or not a
        new connection was required (True or False) or None if an exception
        passed silently.
        """
        if self.mailbox:
            # Nothing to do if the mailbox is already open.
            return False

        credentials = (self.client_id, self.client_secret)
        account = O365.Account(credentials, auth_flow_type='credentials', tenant_id=self.tenant_id)
        self.log_handler.flush()
        try:
            if account.authenticate():
                kwargs = settings.O365_MAIL_MAILBOX_KWARGS
                self.mailbox = account.mailbox(**kwargs)
                return True
            else:
                msg = self.log_handler.get_message()
                if msg:
                    raise Exception(msg)
        except Exception as e:
            if not self.fail_silently:
                raise

    def close(self):
        """
        As far as I know, the O365 Python API has no method to close the connection.
        """
        pass

    def send_messages(self, email_messages):
        if not email_messages:
            return 0
        with self._lock:
            new_mailbox_created = self.open()
            if not self.mailbox or new_mailbox_created is None:
                return 0
            num_sent = 0
            for message in email_messages:
                sent = self._send(message)
                if sent:
                    num_sent += 1
            if new_mailbox_created:
                self.close()
        return num_sent

    def _send(self, email_message):
        """A helper method that does the actual sending."""
        if not email_message.recipients():
            return False

        # Basic email information
        m = self.mailbox.new_message()
        m.to.add(email_message.to)
        m.cc.add(email_message.cc)
        m.bcc.add(email_message.bcc)

        m.sender.name, m.sender.address = util.get_name_and_email(email_message.from_email)
        m.subject = email_message.subject
        m.body = util.get_message_body(email_message)
        
        # Attachments
        if email_message.attachments:
            for attachment in email_message.attachments:
                # The attachment can either be a MIME object or a tuple according to Django docs.
                # We need to convert it to a file object for the O365 API!
                converter = util.get_converter(attachment)(attachment) # get_converter returns a reference to a function, thus it's ()()!
                file = converter.get_file()
                filename = converter.get_filename()

                attachment_count = len(m.attachments)
                m.attachments.add([(file, filename)])
                att_obj = m.attachments[attachment_count] # count is +1 compared to index, so we already have the correct index

                # This is to support inline content (e.g. images)
                att_obj.is_inline = converter.is_inline()
                att_obj.content_id = converter.get_content_id()
        
        # Send it!
        try:
            if (settings.DEBUG and settings.O365_ACTUALLY_SEND_IN_DEBUG) or not settings.DEBUG:
                return m.send(save_to_sent_folder=settings.O365_MAIL_SAVE_TO_SENT)
            return True
        except Exception as e:
            if self.fail_silently:
                return False
            raise e




"""
This is essentially an interface for other converter classes,
these methods MUST be present!
"""

class BaseConverter:
    def get_file(self):
        raise NotImplementedError()

    def get_filename(self):
        raise NotImplementedError()

    def is_inline(self):
        raise NotImplementedError()

    def get_content_id(self):
        raise NotImplementedError()


import logging


class SimpleErrorHandler(logging.Handler):
    level = logging.ERROR
    exceptions = None
    
    def __init__(self):
        self.exceptions = []
        super().__init__()
    
    def handle(self, record):
        self.exceptions.append(record.getMessage())
        self.exceptions = self.exceptions[:5]

    def get_message(self):
        message = None
        if self.exceptions:
            message = self.exceptions[-1]
            self.exceptions.remove(message)
        return message
    
    def flush(self):
        self.exceptions = []



from .file_base import BaseConverter
from email.mime.base import MIMEBase
import base64
import re
import io


def get_charset(content_type):
    """
    Function to get charset from content-type string such as
    'text/plain; charset="utf-8"', else it returns ascii
    """
    type = re.search(r'charset=.([A-Za-z0-9-./]+).', content_type)
    return type.group(1) if type else 'ascii'


class MIMEObjectToFileObject(BaseConverter):
    def __init__(self, obj):
        assert isinstance(obj, MIMEBase) == True, "Object must be an instance of MIMEBase!"
        self.obj = obj

    def get_file(self):
        """Returns a file-like object, in this case a BytesIO object."""
        content_type = self.obj.get('Content-Type') # 'text/plain; charset="utf-8"' or 'application/octet-stream' etc

        self.mimetype = re.search(r'^([A-Za-z0-9-./]+)', content_type).group(0)
        self.charset = get_charset(content_type)
        self.encoding = self.obj.get('Content-Transfer-Encoding') # 'base64' or '7bit' etc
        self.content = self.decode_content(self.obj.get_payload())
        self.filename = self.obj.get_filename() or 'untitled'
        
        return io.BytesIO(self.content)

    def decode_content(self, content):
        """Returns a byte object"""
        if self.encoding == 'base64':
            return self.decode_base64(content)
        else:
            return content.encode(self.charset)

    def decode_base64(self, content):
        return base64.b64decode(content.encode(self.charset))

    def get_filename(self):
        return self.filename

    def is_inline(self):
        return (self.get_content_id() is not None)

    def get_content_id(self):
        return self.obj.get('Content-ID')